from typing import Callable, Any
from dataclasses import dataclass


class Command:
    def __init__(
        self,
        name: str,
        _help: str,
        param: str = "",
        param_help: str = "",
        param_type: type = str,
        param_required: bool = False,
    ) -> None:
        self.name = name
        self.help = _help
        self.param = param
        self.param_help = param_help
        self.param_type = param_type
        self.param_required = param_required


class Group:
    def __init__(
        self,
        _id: int,
        name: str,
        path: str,
        description: str,
        visibility: str,
        web_url: str,
        full_path: str,
        parent_id: int,
    ) -> None:
        self.id = _id
        self.name = name
        self.path = path
        self.description = description
        self.visibility = visibility
        self.web_url = web_url
        self.full_path = full_path
        self.parent_id = parent_id

    def __str__(self) -> str:
        return f'Group({self.id}, "{self.full_path}")'

    def __repr__(self) -> str:
        return self.__str__()


class Project:
    def __init__(
        self,
        _id: int,
        description: str,
        web_url: str,
        name: str,
        path_with_namespace: str,
        star_count: int,
        open_issues_count: int,
        namespace_id: int,
        namespace_name: str,
    ) -> None:
        self.id = _id
        self.description = description
        self.web_url = web_url
        self.name = name
        self.path_with_namespace = path_with_namespace
        self.star_count = star_count
        self.open_issues_count = open_issues_count
        self.namespace = {"id": namespace_id, "name": namespace_name}

    def __str__(self) -> str:
        return f'Porject({self.id}, "{self.path_with_namespace}")'

    def __repr__(self) -> str:
        return self.__str__()


@dataclass
class ProjectQuestion:
    message: str
    name: str
    default: str
    help: str
    res_type: type
    validate: Callable[[str], bool] = None
    transform: Callable[[str], bool] = None
    response: Any = None

    def transform_response(self, response: Any) -> None:
        if response is None:
            return
        
        if self.transform is not None:
            self.response = self.transform(response)

    def set_response(self, response: Any) -> bool:
        if not isinstance(response, self.res_type):
            return False

        if response is None:
            self.response = self.default

        self.response = response
        
        if self.validate is not None:
            return self.validate(response)
        
        return True



from collections import defaultdict
from os import name
import re
from ._typing import ProjectQuestion
from .api import get_owned_groups


def get_namespace_id(namespace_name: str) -> int:
    """Get the namespace id to privide to the api

    Args:
        namespace_name (str): Namespace name

    Returns:
        int: Namespace id
    """
    GROUPS = get_owned_groups()

    _id = [group.id for group in GROUPS if group.full_path == namespace_name]

    if len(_id) > 0:
        return _id[0]
    else:
        return -1


QUESTIONS = [
    ProjectQuestion(
        message="💡 Project name",
        name="name",
        default="",
        help="The name cannot contains space",
        res_type=str,
        validate=lambda v: bool(re.search(r"^[\w\d\-\s]*$", v) and len(v) > 0),
    ),
    ProjectQuestion(
        message="📁 Project path (use empty value to skip)",
        name="path",
        default="",
        help="The name cannot contains space",
        res_type=str,
        validate=lambda v: bool(re.search(r"^[\w\d\-]*$", v)),
    ),
    ProjectQuestion(
        message="📄 Project description (use empty value to skip)",
        name="description",
        default="",
        help="The length cannot pass 255",
        res_type=str,
        validate=lambda v: bool(len(v) <= 255),
    ),
    ProjectQuestion(
        message="📖 Initialize project with a README (y, n)",
        name="initialize_with_readme",
        default="n",
        help="Choose between y, and n",
        res_type=str,
        validate=lambda v: bool(v in ["y", "n"]),
        transform=lambda v: True if v == "y" else False,
    ),
    ProjectQuestion(
        message="👀 Project visibility",
        name="visibility",
        default="public",
        help="Choose between public, internal, or private",
        res_type=str,
        validate=lambda v: bool(v in ["public", "internal", "private"]),
    ),
    ProjectQuestion(
        message="🗃️ Project namespace (<group>\[/<subgroup>, ...], use empty value to skip)",
        name="namespace_id",
        default="",
        help="",
        res_type=str,
        validate=lambda v: bool(re.search(r"(^$|^[\w\d.-\/]*$|)", v)),
        transform=lambda v: get_namespace_id(v) if v != "" else None,
    ),
]



#!/usr/bin/python
"""hmp - HEPIA project manager

This script is a helper for all students at HEPIA
to assist the creation of a new coding project with
a gitlab (https://gitedu.hesge.ch) repo directly link
to the local instance.

@author : Tanguy Cavagna <tanguy.cvgn@ik.me>
@copyrigth : Copyright 2021, hpm
@version : 0.1
@status : developpement
"""

from typing import Text

from rich.box import DOUBLE_EDGE
from . import args as arguments
from . import api
from dotenv import load_dotenv
import pathlib
from rich import align, print
from rich.panel import Panel
from rich.console import Group, Console
from rich.columns import Columns
from rich.text import Text
from .namespaces import create_namespace_tree
from ._printing import clr_scr, print_command_header, generate_project_tree, del_last_line
from .questions import QUESTIONS
from .errors import APIError

def main():
    # Append the .env file to the current symlink/file
    env_path = pathlib.Path(__file__).resolve().parent.absolute() / ".env"
    load_dotenv(dotenv_path=env_path)

    args = arguments.parse_args()
    console = Console()

    if args.command == "setup":
        if not args.token:
            print("Command `setup` require argument `--token`.")
            exit()
        
        with open(pathlib.Path(__file__).resolve().parent.absolute() / "config.txt", "w+") as f:
            f.write(args.token)
    elif args.command == "create-app":
        # @see: https://github.com/kefranabg/readme-md-generator for good cli use guidance
        print_command_header(args.command)

        for question in QUESTIONS:
            tmp_res = None
            prompt = f"[green]?[/green] [bright_white]{question.message}[/bright_white] "
            default_indication = f"[dim italic]({question.default})[/dim italic] "
            
            if question.default != "":
                prompt += default_indication
            
            # Prompt the question
            tmp_res = console.input(prompt)

            # Set the default value as current value when no provided
            if question.default != "" and tmp_res == "":
                prompt = prompt.replace(default_indication, "")
                tmp_res = question.default

            # Validate the user input
            while not question.set_response(tmp_res):
                del_last_line()
                tmp_res = console.input(f"{prompt}[red1]({question.help})[/red1] ")

            # Remove default indication prompt when entered a valid input
            if tmp_res != "":
                prompt = prompt.replace(default_indication, "")

            del_last_line()
            print(f"{prompt}[cyan]{question.response}[/cyan]")

            # Transform the response if needed for the api
            if question.transform != None:
                question.transform_response(tmp_res)
        
        # Create the api params list
        api_params = []
        for q in QUESTIONS:
            api_params.append({
                "key": q.name,
                "value": q.response
            })
        
        try:
            new_project = api.create_project(api_params)

            print("✅ Project created")
            print()
            print(f"👀 View the created project on [link {new_project.web_url}]{new_project.web_url}!")
            print()
            print(Panel(Text("Project was successfully created.\nThanks for using hpm !", justify="center"), box=DOUBLE_EDGE, border_style="bright_cyan", padding=(1, 2), expand=False))
            print()
        except APIError as e:
            print(e.message)
    elif args.command == "list":
        page = 1
        prev_page = 0
        co = ""

        # Emulate a `do {...} while()`
        while True:
            clr_scr()
            print_command_header(args.command)

            # Only call the api if the pages have changed
            if prev_page != page:
                groups = api.get_owned_groups()
                owned_projects = api.get_owned_projects(page)

                if groups is None or owned_projects is None:
                    exit()

                groups = sorted(groups, key=lambda k: k.full_path)
                headers = owned_projects["headers"]
                group_projects = owned_projects["list"]

            prev_page = "<" if (headers["X-Prev-Page"] != "") else ""
            next_page = ">" if (headers["X-Next-Page"] != "") else ""

            roots = []
            for root in create_namespace_tree(groups, group_projects):
                roots.append(Panel(generate_project_tree(root)))

            # Format navigation menu
            navigation = Group(
                Panel(
                    Group(
                        Text.assemble(
                            ("Page", "bright_white"),
                            f": {prev_page} ",
                            (str(page), "bright_white"),
                            "/",
                            (headers["X-Total-Pages"], "bright_white"),
                            f" next_page",
                            justify="center",
                        ),
                        Text.assemble(
                            ("Navigation", "bright_white"),
                            ": Previous (",
                            ("p", "green"),
                            ") | Next (",
                            ("n", "green"),
                            ") | Close (",
                            ("c", "red1"),
                            ") | ",
                            justify="center",
                        ),
                    )
                )
            )

            # Print
            print(Columns(roots))
            print(navigation)

            co = console.input("[bold magenta]Action: [/bold magenta]")

            # List commands
            if co == "c":
                break
            elif co == "p":
                if headers["X-Prev-Page"] != "":
                    prev_page = page
                    page = headers["X-Prev-Page"]
            elif co == "n":
                if headers["X-Next-Page"] != "":
                    prev_page = page
                    page = headers["X-Next-Page"]
            else:
                prev_page = page
                continue

            # `do {...} while()` condition
            if co == "c":
                break
    # Exit when command unknown
    else:
        exit()

if __name__ == "__main__":
    main()


import argparse
from ._typing import Command

COMMANDS = [
    Command("create-app", "Create a new project."),
    Command("list", "List all your projects."),
    Command("setup", "Save API token.", param="token", param_help="Store API token", param_type=str),
]


def parse_args():
    parser = argparse.ArgumentParser(description="HEPIA project manager.")

    subparsers = parser.add_subparsers(help="Custom commands", dest="command")

    for c in COMMANDS:
        tmp_parser = subparsers.add_parser(c.name, help=c.help)
        if c.param != "":
            tmp_parser.add_argument(f"--{c.param}", help=c.param_help, type=c.param_type, required=c.param_required)

    args = parser.parse_args()

    args_values = vars(args)
    if not any(args_values.values()):
        parser.error("No commands provided.")

    return args



import requests
import json
import pathlib
from urllib.parse import urlencode
from typing import List, Union
from rich import print
from rich.console import Console
from ._typing import Group, Project
from .errors import APIError


BASE_URI = "https://gitedu.hesge.ch/api/v4"
console = Console()

try:
    token = open(pathlib.Path(__file__).resolve().parent.absolute() / "config.txt", "r").read()
except Exception as e:
    token = ""


def get_owned_groups() -> List[Group]:
    """Get user owned groups

    Returns:
        Group[]: List of owned groups
    """
    if token == "":
        print("No token as been register. Please do so using `setup --token <TOKEN_VALUE>.`")
        return None

    url = f'{BASE_URI}/groups?owned=true&per_page=100'

    payload = {}
    headers = {
        'Private-Token': token
    }

    with console.status("[bold green]💾 Getting groups...") as status:
        response = requests.request("GET", url, headers=headers, data=payload).text
    response = json.loads(response)

    groups: List[Group] = []
    # @see: https://gitedu.hesge.ch/help/api/groups.md#list-groups
    for g in response:
        groups.append(Group(g['id'], g['name'], g['path'], g['description'],
                      g['visibility'], g['web_url'], g['full_path'], g['parent_id']))

    return groups


def get_owned_projects(page=1):
    """Get user owned projects

    Return:
        Project[]: List of owned projects
    """
    if token == "":
        print("No token as been register. Please do so using `setup --token <TOKEN_VALUE>.`")
        return None

    url = f'{BASE_URI}/projects?membership=true&page={page}'

    payload = {}
    headers = {
        'Private-Token': token
    }

    with console.status("[bold green]💾 Getting projects...") as status:
        response = requests.request("GET", url, headers=headers, data=payload)
    response_json = json.loads(response.text)

    p_list: List[Project] = []
    # @see: https://gitedu.hesge.ch/help/api/projects.md#list-all-projects
    for p in response_json:
        p_list.append(Project(p['id'], p['description'], p['web_url'], p['name'],
                              p['path_with_namespace'], p['star_count'], p['open_issues_count'], p['namespace']['id'], p['namespace']['name']))

    return {
        # @see: https://stackoverflow.com/a/45087988 for informations on headers
        'headers': {
            'X-Next-Page': response.headers.get('X-Next-Page'),
            'X-Prev-Page': response.headers.get('X-Prev-Page'),
            'X-Total-Pages': response.headers.get('X-Total-Pages'),
        },
        # Sort the project by full namespace for easier print later
        'list': sorted(p_list, key=lambda k: k.path_with_namespace)
    }


def create_project(params:list) -> Union[Project, APIError]:
    """Create a new project

    Args:
        params (list): Param list

    Returns:
        Project: Newly created project
    """
    if token == "":
        print("No token as been register. Please do so using `setup --token <TOKEN_VALUE>.`")
        return None

    url = f'{BASE_URI}/projects?'

    # Build url parameters
    url_params = {}
    for param in params:
        if param["value"] is not None and param["value"] != "":
            url_params[param["key"]] = param["value"]

        # Special case for namespace check
        if param["key"] == "namespace_id" and param["value"] == -1:
            return APIError(message="Namespace not found")

    url += urlencode(url_params)

    payload = {}
    headers = {
        'Private-Token': token
    }

    with console.status("[bold green]🚀 Creating project...") as status:
        response = requests.request("POST", url, headers=headers, data=payload)
    response_json = json.loads(response.text)

    return Project(
        response_json['id'],
        response_json['description'],
        response_json['web_url'],
        response_json['name'],
        response_json['path_with_namespace'],
        response_json['star_count'],
        response_json['open_issues_count'],
        response_json['namespace']['id'],
        response_json['namespace']['name']
    )
        


import os
import platform
import sys
from datetime import datetime
from rich import print
from rich.tree import Tree
from rich.panel import Panel
from rich.columns import Columns
from rich.text import Text

CURSOR_UP_ONE = "\x1b[1A"
ERASE_LINE = "\x1b[2K"


def generate_project_tree(namespace: dict, tree: Tree = None) -> None:
    """Print the project list

    Args:
        namespace (dict): Current namespace
        tree (Tree): Tree to build
    """
    if tree is None:
        tree = Tree(
            f'[link {namespace["url"]}][bold cyan]{namespace["name"]}[/bold cyan][/link {namespace["url"]}]'
        )
    else:
        tree = tree.add(
            f'[link {namespace["url"]}][bold cyan]{namespace["name"]}[/bold cyan][/link {namespace["url"]}]'
        )

    # Recursivly print the children
    for c in namespace["children"]:
        generate_project_tree(c, tree=tree)

    # Print the projects list
    for i, p in enumerate(namespace["projects"]):
        tree.add(
            f"{p.name} :glowing_star:{p.star_count} :name_badge:{p.open_issues_count}"
        )

    return tree


def print_command_header(command: str):
    """Global header

    Args:
        command (str): Executed command
    """
    print(
        Panel(
            Columns(
                [
                    Text("Made By Tanguy Cavagna with 💖", style="italic"),
                    Text(
                        datetime.now().strftime("%Y-%m-%d, %H:%M:%S"),
                        justify="right",
                        style="italic",
                    ),
                ],
                expand=True,
                title="[bold bright_white]HES-SO[/bold bright_white] - [bold red1]HEPIA[/bold red1]",
            ),
            border_style="bright_white",
            title=f"Command: [orange_red1]{command}[/orange_red1]",
            title_align="left",
        )
    )


def clr_scr():
    """Clear screen on all os"""
    if platform.system().lower() == "windows":
        os.system("cls")
    else:
        os.system("clear")


def del_last_line():
    """Delete the last console line"""
    sys.stdout.write(CURSOR_UP_ONE)
    sys.stdout.write(ERASE_LINE)
